"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the routes for the minion-ci API.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import re
import mongoengine
from flask import Blueprint, request, jsonify, render_template

from .models import Job, mongo_to_dict
from .core import workers

api = Blueprint("api", __name__)


@api.route("/", methods=["GET"])
def index():
    """Serve index HTML."""
    return render_template("index.html", jobs=Job.objects)


@api.route("/status", methods=["GET"])
def get_status():
    """Get status of the minion server."""
    status = {
        "queue": {
            "size": workers.queue.qsize()
        },
        "jobs": {

        }
    }
    return jsonify(status)


@api.route("/jobs", methods=["GET"])
def get_jobs():
    """Serve list of jobs."""
    return jsonify({"jobs": [mongo_to_dict(j) for j in Job.objects]})


@api.route("/jobs/<job_id>", methods=["GET"])
def get_job(job_id):
    """Serve information of specific job."""
    job = Job.objects.get(id=job_id)
    return jsonify({"job": mongo_to_dict(job)})


@api.route("/jobs", methods=["POST", "PUT"])
def create_job():
    """Create a new job"""
    data = request.json
    job = {
        "repository_url": data.pop("repo_url", None),
        "commit_hash": data.pop("commit_hash", None),
        "branch": data.pop("branch", None),
        "keep_data": data.pop("keep_data", False),
        "attributes": data
    }
    workers.queue.put(job)
    return jsonify({"status": True})


@api.route("/jobs", methods=["DELETE"])
def delete_jobs():
    """Delete all jobs."""
    Job.drop_collection()
    return jsonify({"status": True})


@api.route("/jobs/<job_id>", methods=["DELETE"])
def delete_job(job_id):
    """Delete a specific job."""
    try:
        job = Job.objects.get(id=job_id)
    except mongoengine.errors.ValidationError:
        return jsonify({"status": False, "msg": "No such job id found: {0}".format(job_id)})
    else:
        job.delete()
        return jsonify({"status": True})



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the core functionality for the minion-ci client.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""


import requests
import subprocess
from time import sleep


def ensure_started_server(func):
    """Ensures that the minion-ci server is started."""
    def _wrapper(self, *args, **kwargs):
        """decorator wrapper"""
        try:
            self.get_status()
        except requests.exceptions.ConnectionError:  # server is not started
            subprocess.Popen(["minion-server"])
            sleep(2)

        return func(self, *args, **kwargs)

    return _wrapper


class Client:
    """Client for the minion-ci server."""
    def __init__(self, host, port, schema="http://"):
        self.host = host
        self.port = port
        self.base_url = "{0}{1}:{2}".format(schema, host, port)

    def _build_url(self, route):
        """Build the API URL of the given route."""
        return "{0}/{1}".format(self.base_url, route)

    def get_status(self):
        """Get status of the minion server."""
        response = requests.get(self._build_url("status"))
        return response.json()

    @ensure_started_server
    def get_jobs(self):
        """Get all jobs from the minion server."""
        response = requests.get(self._build_url("jobs"))
        return response.json()

    @ensure_started_server
    def get_job(self, job_id):
        """Get a single job from the minion server."""
        response = requests.get(self._build_url("jobs/{0}".format(job_id)))
        return response.json()

    @ensure_started_server
    def submit(self, repository_url, commit_hash=None, branch=None, keep_data=None, attributes=None):
        """Submit a new job to the minion server."""
        data = {
            "repo_url": repository_url,
            "commit_hash": commit_hash,
            "branch": branch,
            "keep_data": keep_data,
        }
        if attributes:
            data.update(attributes)
        response = requests.put(self._build_url("jobs"), json=data)
        return response.json()



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module launches the minion client.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import os
import json
import click
import shutil
import subprocess

from .core import Client


@click.group()
@click.pass_context
@click.version_option()
def cli(ctx):
    """Use your minion to test your code."""
    ctx.obj = Client("127.0.0.1", 5000)


@cli.command()
def init():
    """Initialize minion aliases for git repository."""
    # check if git repository
    if subprocess.call(["git", "status"], stdout=subprocess.PIPE, stderr=subprocess.PIPE) != 0:
        click.echo("This is not a git repository!")
        return 1

    if os.path.exists("minion.yml"):
        click.echo("This is already a minion repository!")
        return

    if subprocess.call(["git", "config", "alias.minion", "!minion submit ."]) != 0:
        click.echo("Unable to add git minion alias!")
        return 1

    shutil.copyfile(os.path.join(os.path.dirname(__file__), "templates/minion.yml"), "minion.yml")
    click.echo("Added 'git minion' alias")


@cli.command()
@click.option("--json", "as_json", flag_value=True,
              help="Output jobs as JSON.")
@click.pass_context
def jobs(ctx, as_json):
    """Get all jobs from the minion server."""
    jobs = ctx.obj.get_jobs()
    if as_json:
        click.echo(json.dumps(jobs, indent=4, sort_keys=True))
    else:
        for job in jobs["jobs"]:
            click.echo("Job {0}:".format(click.style(job["id"], bold=True)))
            click.echo("    Created at: {0}".format(
                click.style(job["created_at"], bold=True)))
            click.echo("    Repository URL: {0}".format(
                click.style(job["repository_url"], bold=True)))
            click.echo("    Commit Hash: {0}".format(
                click.style(job["commit_hash"], bold=True)))
            click.echo("    Branch: {0}".format(
                click.style(job["branch"], bold=True)))
            click.echo("    Status: {0}".format(
                click.style("✓ success" if job["result"]["status"] else "✘ failure", bold=True,
                            fg="green" if job["result"]["status"] else "red")))
            if not job["result"]["status"]:
                click.echo("        Error Message: {0}".format(
                    click.style(job["result"]["error_msg"], fg="red")))

            click.echo()


@cli.command()
@click.argument("job_id")
@click.option("--json", "as_json", flag_value=True,
              help="Output jobs as JSON.")
@click.pass_context
def job(ctx, job_id, as_json):
    """Get information about a single job from the minion server."""
    job = ctx.obj.get_job(job_id)
    if as_json:
        click.echo(json.dumps(job, indent=4, sort_keys=True))
    else:
        job_data = job["job"]
        click.echo("Id: {0}".format(click.style(job_data["id"], bold=True)))
        click.echo("Created at: {0}".format(
            click.style(job_data["created_at"], bold=True)))
        click.echo("Repository URL: {0}".format(
            click.style(job_data["repository_url"], bold=True)))
        click.echo("Commit Hash: {0}".format(
            click.style(job_data["commit_hash"], bold=True)))
        click.echo("Branch: {0}".format(
            click.style(job_data["branch"], bold=True)))
        click.echo("Status: {0}".format(
            click.style("✓ success" if job_data["result"]["status"] else "✘ failure", bold=True,
                        fg="green" if job_data["result"]["status"] else "red")))
        if not job_data["result"]["status"]:
            click.echo("    Error Message: {0}".format(
                click.style(job_data["result"]["error_msg"], fg="red")))


@cli.command()
@click.argument("repo_url")
@click.option("-h", "--commit-hash",
              help="The git commit hash to test.")
@click.option("-b", "--branch",
              help="The git branch to test.")
@click.option("-k", "--keep-data", flag_value=True,
              help="If given the result data will not be deleted.")
@click.option("-a", "--attribute", "raw_attributes", multiple=True,
              type=(str, str),
              help="Specify arbitrary attributes.")
@click.pass_context
def submit(ctx, repo_url, commit_hash, branch, keep_data, raw_attributes):
    """Submit a new job."""
    # check if given repo url is path or url
    if os.path.exists(repo_url):  # is local path -> get absolute
        repo_url = os.path.abspath(repo_url)

    attributes = {a[0]: a[1] for a in raw_attributes}
    response = ctx.obj.submit(repo_url, commit_hash, branch, keep_data, attributes)
    if response["status"]:
        click.echo("Submitted new job - Stay tuned for results!")
    else:
        click.echo("Something failed while submitting the job!")


main = cli

if __name__ == "__main__":
    main()



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the models to persist CI job data.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import os
import datetime
from flask.ext.mongoengine import MongoEngine

db = MongoEngine()


class Result(db.EmbeddedDocument):  # pylint: disable=no-init,too-few-public-methods
    """Document to store a job test result."""
    created_at = db.DateTimeField(default=datetime.datetime.now, required=True)
    status = db.BooleanField(default=False, required=True)
    error_msg = db.StringField()
    logs = db.StringField()


class Job(db.Document):  # pylint: disable=no-init,too-few-public-methods
    """Document to store a single minion-ci job."""
    meta = {
                'ordering': ['-created_at']
    }

    created_at = db.DateTimeField(default=datetime.datetime.now, required=True)
    repository_url = db.StringField(required=True)
    commit_hash = db.StringField()
    branch = db.StringField()
    attributes = db.DictField()
    result = db.EmbeddedDocumentField(Result)

    config_file = "minion.yml"


def mongo_to_dict(obj):
    """Turn a mongoengine object into a jsonible python object."""

    return_data = []
    if isinstance(obj, db.Document):
        return_data.append(("id", str(obj.id)))

    for field_name in obj._fields:
        data = obj._data[field_name]

        if isinstance(obj._fields[field_name], db.DateTimeField):
            return_data.append((field_name, str(data.isoformat())))
        elif isinstance(obj._fields[field_name], db.StringField):
            return_data.append((field_name, str(data)))
        elif isinstance(obj._fields[field_name], db.BooleanField):
            return_data.append((field_name, bool(data)))
        elif isinstance(obj._fields[field_name], db.FloatField):
            return_data.append((field_name, float(data)))
        elif isinstance(obj._fields[field_name], db.IntField):
            return_data.append((field_name, int(data)))
        elif isinstance(obj._fields[field_name], (db.ListField, db.DictField)):
            return_data.append((field_name, data))
        elif isinstance(obj._fields[field_name], db.EmbeddedDocumentField):
            return_data.append((field_name, mongo_to_dict(data)))
        elif isinstance(obj._fields[field_name], db.ObjectIdField):
            return_data.append((field_name, str(data)))

    return dict(return_data)



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

__VERSION__ = "0.0.2"



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the server code for `minion-ci`.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import re
import click
import yaml
from flask import Flask
from jinja2 import evalcontextfilter, Markup, escape

from .config import DefaultConfig
from .routes import api
from .models import db
from .core import workers


def parse_config(path):
    """Parse minion-ci server configuration file."""
    with open(path) as config_file:
        return yaml.load(config_file)


@evalcontextfilter
def nl2br(eval_ctx, value):
    _paragraph_re = re.compile(r'(?:\r\n|\r|\n){2,}')
    result = u'\n\n'.join(u'<p>%s</p>' % p.replace('\n', '<br>\n') \
        for p in _paragraph_re.split(escape(value)))
    if eval_ctx.autoescape:
        result = Markup(result)
    return result


def create_app(name):
    """Create flask app"""
    app = Flask(__name__)
    app.register_blueprint(api)
    app.config.from_object(DefaultConfig)

    try:
        app.config.update(parse_config(app.config["DEFAULT_CONFIGURATION_FILE"]))
    except FileNotFoundError:
        pass

    # add required filters
    app.jinja_env.filters["nl2br"] = nl2br

    # initialize mongodb engine for use in flask
    db.init_app(app)

    # initilize worker pool and job queue
    workers.init_app(app)

    return app



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the exceptions specific to `minion` errors.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

class MinionError(Exception):
    """Exception which is raised for minion specific errors."""
    pass



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This package contains the minion-ci server.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the parser to parse the `minion.yml` file.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import yaml

from .errors import MinionError


def parse(path):
    """Parse the given minion.yml file"""
    try:
        with open(path) as minion_file:
            config = yaml.load(minion_file)
    except OSError:
        raise MinionError("No minion.yml config file found in repository")
    return config



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the core functionality for the minion-ci server.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import os
import shutil
from flask import current_app
from multiprocessing import Pool, Queue
from subprocess import Popen, PIPE, STDOUT

from .models import Job, Result
from ..parser import parse
from ..errors import MinionError


class WorkersExtension:
    """A flask extension for the job queue and worker pool."""
    def __init__(self, app=None):
        self._app_cache = {}

        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        queue = Queue()
        pool = Pool(1, worker, (queue, app.config))
        self._app_cache[app] = {
            "queue": queue,
            "pool": pool
        }

    @property
    def queue(self):
        """Return the job queue used by the workers."""
        return self._app_cache[current_app._get_current_object()]["queue"]

    @property
    def pool(self):
        """Return the worker pool."""
        return self._app_cache[current_app._get_current_object()]["pool"]

workers = WorkersExtension()


def worker(queue, config):
    """Single worker to process jobs from the given queue."""
    while True:
        job_data = queue.get(True)

        # insert job into database
        job = Job()
        job.repository_url = job_data["repository_url"]
        job.commit_hash = job_data["commit_hash"]
        job.branch = job_data["branch"]
        job.attributes = job_data["attributes"]
        job.save()

        process(job, config, job_data["keep_data"])


def process(job, config, keep_data):
    """
        Process the given job.

        The process is composed of the following steps:
            - clone git repository
            - run before run command
            - run command
            - run after run command
    """
    result = Result()
    logs = ""

    local_repo_path = os.path.join(config["JOB_DATAPATH"], str(job.id))

    try:
        # clone the git repository
        git_clone = Popen(["git", "clone", job.repository_url, local_repo_path], stdout=PIPE, stderr=STDOUT)
        tmp_stdout, _ = git_clone.communicate()
        logs += tmp_stdout.decode("utf-8")
        if git_clone.returncode != 0:
            raise MinionError("Failed to clone git repository from {0} to {1}".format(
                job.repository_url, local_repo_path))

        if job.commit_hash:
            git_reset = Popen(["git", "reset", "--hard", job.commit_hash],
                              cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = git_reset.communicate()
            logs += tmp_stdout.decode("utf-8")
            if git_reset.returncode != 0:
                raise MinionError("Failed to reset repository to {0}".format(job.commit_hash))
        elif job.branch:
            git_checkout = Popen(["git", "checkout", job.branch],
                                 cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = git_checkout.communicate()
            logs += tmp_stdout.decode("utf-8")
            if git_checkout.returncode != 0:
                raise MinionError("Failed to checkout branch {0}".format(job.branch))

        # get current git commit hash
        git_rev_parse = Popen(["git", "rev-parse", "HEAD"],
                              cwd=local_repo_path, stdout=PIPE, stderr=PIPE)
        current_commit_hash, _ = git_rev_parse.communicate()
        job.commit_hash = current_commit_hash.decode("utf-8").strip()
        if git_rev_parse.returncode != 0:
            raise MinionError("Failed to parse current git revision")

        config = parse(os.path.join(local_repo_path, job.config_file))

        if "precondition" in config:
            before_run = Popen(config["precondition"], shell=True,
                               cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = before_run.communicate()
            logs += tmp_stdout.decode("utf-8")
            if before_run.returncode != 0:
                raise MinionError("Failed to run precondition command: '{0}'".format(config["precondition"]))

        if "command" in config:
            command_run = Popen(config["command"], shell=True,
                                cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = command_run.communicate()
            logs += tmp_stdout.decode("utf-8")
            if command_run.returncode != 0:
                raise MinionError("Failed to run command: '{0}'".format(config["command"]))
    except MinionError as e:
        result.error_msg = str(e)
        result.status = False
        if True in config and "failure" in config[True]:
            after_run = Popen(config[True]["failure"], shell=True,
                              cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = after_run.communicate()
            logs += tmp_stdout.decode("utf-8")
            if after_run.returncode != 0:
                result.error_msg += "\n{0}".format("Failed to run failure command: '{0}'".format(
                    config[True]["failure"]))
    else:
        result.status = True
        if True in config and "success" in config[True]:
            after_run = Popen(config[True]["success"], shell=True,
                              cwd=local_repo_path, stdout=PIPE, stderr=STDOUT)
            tmp_stdout, _ = after_run.communicate()
            logs += tmp_stdout.decode("utf-8")
            if after_run.returncode != 0:
                result.error_msg = "Failed to run failure command: '{0}'".format(
                    config[True]["failure"])
                result.status = False
    finally:
        result.logs = logs

        job.result = result
        job.save()

        if not keep_data:
            shutil.rmtree(local_repo_path)



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module launches the minion-ci server.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import click
import logging

@click.command()
@click.option("-c", "--config",
              help="Path to configuration file.")
@click.version_option()
def main(config):
    """Manage minion-ci server."""
    from .app import create_app, parse_config

    app = create_app("minion")
    logging.basicConfig(filename=app.config["LOG_FILE_PATH"], level=logging.DEBUG)

    if config:
        app.config.update(parse_config(config))

    app.run(debug=True)


if __name__ == "__main__":
    main()



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains helper functions for the minion-ci server and client

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import os
from urllib.parse import urlparse

def get_repository_name(repository_url):
    """
        Returns the repository name from the
        given repository_url.
    """
    parts = urlparse(repository_url)
    return os.path.splitext(os.path.basename(parts.path))[0]



"""
    `minion-ci` is a minimalist, decentralized, flexible Continuous Integration Server for hackers.

    This module contains the server config for minion-ci.

    :copyright: (c) by Timo Furrer
    :license: MIT, see LICENSE for details
"""

import os
import click

class DefaultConfig:
    # default configuration settings
    MONGODB_SETTINGS = {
        "db": "minion-ci",
        "connect": False  # lazy connect
    }
    APPLICATION_DATAPATH = click.get_app_dir("minion-ci")
    JOB_DATAPATH = os.path.join(APPLICATION_DATAPATH, "jobs")
    DEFAULT_CONFIGURATION_FILE = os.path.join(APPLICATION_DATAPATH, "server-config.yml")
    LOG_FILE_PATH = os.path.join(APPLICATION_DATAPATH, "server.log")



# -*- coding: utf-8 -*-

"""
    Setup minion-ci package.

    Only support Python versions > 3.4.1
"""

import ast
import re
import sys

from setuptools import setup, find_packages
from pip.req import parse_requirements

if sys.version_info < (3, 4, 1):
    raise RuntimeError("minion-ci requires Python 3.4.1+")


def get_version():
    """Gets the current version"""
    _version_re = re.compile(r"__VERSION__\s+=\s+(.*)")
    with open("minion/__init__.py", "rb") as init_file:
        version = str(ast.literal_eval(_version_re.search(
            init_file.read().decode("utf-8")).group(1)))
    return version


setup(
    name="minion-ci",
    version=get_version(),
    license="MIT",

    description="minimalist, decentralized, flexible Continuous Integration Server for hackers",
    long_description=open("./README.rst", "r", encoding="utf-8").read(),

    author="Timo Furrer",
    author_email="tuxtimo@gmail.com",

    url="https://github.com/timofurrer/minion-ci",

    packages=find_packages(),
    include_package_data=True,

    install_requires=list(x.name for x in parse_requirements("./requirements.txt")),

    entry_points={
        "console_scripts": [
            "minion-server=minion.server.__main__:main",
            "minion=minion.client.__main__:main"
        ]
    },

    keywords=[
        "minion-ci",
        "minimalist", "simple", "flexible", "decentral",
        "ci", "continuous integration", "server",
        "git", "github",
        "source", "code",
        "docker", "container",
    ],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Operating System :: OS Independent",
        "Environment :: Console",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: End Users/Desktop",
        "Topic :: Utilities",
    ],
)
